// DebugUtility.h: Debug Utility functions/macros.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DEBUGUTILITY_H__485D8D51_3CF5_467F_9C77_07C9AB80E5D4__INCLUDED_)
#define AFX_DEBUGUTILITY_H__485D8D51_3CF5_467F_9C77_07C9AB80E5D4__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


// if not using MFC, support for ASSERTs and TRACE is turned off,
//
// emulate it
#ifndef ASSERT

	#ifdef _DEBUG

		#include <crtdbg.h>
		#define ASSERT _ASSERTE 
		
		#define TRACE( strMessage) \
			_CrtDbgReport( _CRT_WARN, NULL, 0, NULL, "%s\n", strMessage)
			// _CrtDbgReport( _CRT_WARN, __FILE__, __LINE__, NULL, "%s\n", strMessage)

	#else
	
		#define ASSERT( bShouldBeTrue) ( ( void)0)
		#define TRACE( strMessage) ( ( void)0)
	#endif
#endif


/*
	used to remind that certain functions/ portions of
	code need to be implemented

	Use this instead of something like ASSERT( FALSE), or
	something similar, 
	since it's very easy to find within code
	(just search for 'IMPLEMENT_ME')
*/
#define IMPLEMENT_ME ASSERT( FALSE);



/*
	acts like an ASSERT in DEBUG versions

	when on release, it still executes
	the instructions that it needs to validate

	(see MFC documentation for more details)
*/
#ifndef VERIFY

	#ifdef _DEBUG

		#define VERIFY( x) ASSERT( x)
	#else

		#define VERIFY( x) ( x)
	#endif // _DEBUG


#endif // VERIFY




/*
	put this into the sections of code 
	you have implemented, but could not test

	This way, you will know if you run into 
	a piece of code that was not tested yet.

	After testing the code, you should remove
	this ('NOT_TESTED') from the code
*/
#define NOT_TESTED ASSERT( FALSE);



// CHG JT, 26.03.2001
// Description
// 
//		traces every element of an STL array
//		only useful for debugging
// 
// Parameters
// const Cont & Array
// Return Value
// 
// Remarks
// 
//		
template< class Cont>
	void TRACE_STL_ARRAY( const Cont & Array)
{
	typename Cont::const_iterator 
		itFirst = Array.begin(),
		itLast = Array.end();

	while ( itFirst != itLast)
	{
		TRACE( to_string( *itFirst).c_str() );
		++ itFirst;
	}
}


#endif // !defined(AFX_DEBUGUTILITY_H__485D8D51_3CF5_467F_9C77_07C9AB80E5D4__INCLUDED_)
