/**********************************************************************
*
COPYRIGHT (C) 2002 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: default_args.hpp
//
// Author: Thomas Becker
//
// Created: September 2002
//
// Description: Class templates and C++ template metafunctions
//              for generalized default function arguments.
//

// Multiple inclusion protection
//
#ifndef __DEFAULT_ARGS_INCLUDED__
#define __DEFAULT_ARGS_INCLUDED__

#include<boost/type_traits/transform_traits.hpp>
#include<boost/type_traits/cv_traits.hpp>
#include<boost/type_traits/composite_traits.hpp>

namespace TMB
{
  // Meta C++ function to retrieve the n-th type from a tuple.
  // =========================================================
  template<typename FP, int n>
  class nth_parameter_type
  {
  public:
    typedef nth_parameter_type<FP::nested_type, n-1>::ret ret;
  };

  template<typename FP>
  class nth_parameter_type<FP, 0>
  {
  public:
    typedef typename FP::element_type ret;
  };

  // Tag classes to indicate mandatory and default parameters.
  // =========================================================
  class default_parameter_tag{};
  class mandatory_parameter_tag{};

  // Nil class for class client_parameter_tuple
  // ==========================================
  class client_nil
  {
  public:
    typedef client_nil element_type;
    typedef client_nil nested_type;
    client_nil(){}
    client_nil(
      const client_nil&
      ) {}
    client_nil(
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil(
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 4
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 5
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 6
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 7
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 8
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 9
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 10
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 11
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 12
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 13
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 14
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 15
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 16
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      ) {}
    client_nil( // 17
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&, 
      const client_nil&
      /* EXTEND */
      ) {}
  };
  
  // Class client_parameter_tuple
  // ============================
  
  template<typename Head, typename Tail>
  class client_parameter_tuple
  {
    // Typedefs and Friends
    // ====================
    
    typedef typename Head element_type;
    typedef typename Tail nested_type;
    
    typedef boost::add_reference<element_type>::type element_reference_type;
    typedef boost::add_reference<boost::add_const<element_type>::type>::type element_const_reference_type;
    typedef boost::add_reference<nested_type>::type rest_reference_type;

    // Nested Classes
    // ==============
    template<int n>
    class nth_parameter_tag{};

  public:

    // Public Construction/Destruction
    // ===============================
    
    // Constructor from n elements
    client_parameter_tuple(
      element_const_reference_type elem_0,
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 0>::ret>::type>::type elem_1 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 1>::ret>::type>::type elem_2 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 2>::ret>::type>::type elem_3 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 3>::ret>::type>::type elem_4 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 4>::ret>::type>::type elem_5 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 5>::ret>::type>::type elem_6 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 6>::ret>::type>::type elem_7 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 7>::ret>::type>::type elem_8 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 8>::ret>::type>::type elem_9 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 9>::ret>::type>::type elem_10 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 10>::ret>::type>::type elem_11 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 11>::ret>::type>::type elem_12 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 12>::ret>::type>::type elem_13 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 13>::ret>::type>::type elem_14 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 14>::ret>::type>::type elem_15 = client_nil(),
      typename boost::add_reference<boost::add_const<nth_parameter_type<Tail, 15>::ret>::type>::type elem_16 = client_nil()
      /* EXTEND */
      ) :
      m_element(elem_0),
      m_rest(
       elem_1, 
       elem_2,
       elem_3,
       elem_4,
       elem_5,
       elem_6,
       elem_7,
       elem_8,
       elem_9,
       elem_10,
       elem_11,
       elem_12,
       elem_13,
       elem_14,
       elem_15,
       elem_16
       /* EXTEND */
       )
    {
    }

    // Methods
    // =======
    
    // Get functions to retrieve the element and the
    // tail.
    element_reference_type get_element()
    { return m_element; }
    //
    rest_reference_type get_rest()
    { return m_rest; }
    
  private:
    // First element and nested object containing the remaining
    // elements.
    element_type m_element;
    nested_type m_rest;

    // Assigning and copying these babies is bad news.
    client_parameter_tuple(const client_parameter_tuple&);
    client_parameter_tuple& operator=(const client_parameter_tuple&);

  };

  // client_parameter_tuple type generator
  // =====================================
  
  template<
   typename T1=client_nil, 
   typename T2=client_nil, 
   typename T3=client_nil,
   typename T4=client_nil,
   typename T5=client_nil,
   typename T6=client_nil,
   typename T7=client_nil,
   typename T8=client_nil,
   typename T9=client_nil,
   typename T10=client_nil,
   typename T11=client_nil,
   typename T12=client_nil,
   typename T13=client_nil,
   typename T14=client_nil,
   typename T15=client_nil,
   typename T16=client_nil,
   typename T17=client_nil
  >
  class make_client_parameter_tuple_type
  {
  public:
    typedef typename client_parameter_tuple<T1, typename make_client_parameter_tuple_type<
      T2, 
      T3, 
      T4, 
      T5, 
      T6, 
      T7, 
      T8, 
      T9, 
      T10, 
      T11, 
      T12, 
      T13, 
      T14, 
      T15, 
      T16, 
      T17 
      /* EXTEND */ 
      >::ret> ret;
  };

  // Kicks in for zero arguments passed
  template<>
  class make_client_parameter_tuple_type<
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil, 
    client_nil,
    client_nil, 
    client_nil
    /* EXTEND */
    >
  {
    public:
      typedef typename client_nil ret;
  };

  // Nil class for class function_parameter_tuple
  // ============================================
  class function_nil
  {
  public:
    function_nil(client_nil, client_nil) {}
  };
  
  // Class function_parameter_tuple
  // ==============================
  
  template<typename Head, typename Tail>
  class function_parameter_tuple
  {
    // Typedefs and Friends
    // ====================
    
    typedef typename Head element_type;
    typedef typename Tail nested_type;
    
    typedef boost::add_reference<element_type>::type element_reference_type;
    typedef boost::add_reference<boost::add_const<element_type>::type>::type element_const_reference_type;
    typedef boost::add_reference<boost::add_const<nested_type>::type>::type rest_const_reference_type;

    // Nested Classes
    // ==============
    template<int n>
    class nth_parameter_tag{};

  public:

    // Public Construction/Destruction
    // ===============================
    
    // Copy constructor from two client parameter tuple objects, 
    // default implementation: copy element from first argument.
    template<typename Rhs1Head, typename Rhs1Tail, typename Rhs2Head, typename Rhs2Tail>
    function_parameter_tuple(
      const client_parameter_tuple<Rhs1Head, Rhs1Tail>& rhs1,
      client_parameter_tuple<Rhs2Head, Rhs2Tail>& rhs2
      ) :
      m_element(const_cast<client_parameter_tuple<Rhs1Head, Rhs1Tail>&>(rhs1).get_element()),
      m_rest(
        const_cast<client_parameter_tuple<Rhs1Head, Rhs1Tail>&>(rhs1).get_rest(), 
        rhs2.get_rest()
        )
    {
    }

    // Copy constructor from two client parameter tuple  objects, 
    // partial specialization when the element type of the first
    // argument is default_parameter: copy element from second argument.
    template<typename Rhs1Tail, typename Rhs2Head, typename Rhs2Tail>
    function_parameter_tuple(
      const client_parameter_tuple<default_parameter_tag, Rhs1Tail>& rhs1,
      client_parameter_tuple<Rhs2Head, Rhs2Tail>& rhs2) :
      m_element(rhs2.get_element()),
      m_rest(
        const_cast<client_parameter_tuple<default_parameter_tag, Rhs1Tail>&>(rhs1).get_rest(), 
        rhs2.get_rest()
        )
    {
    }
    
    // Methods
    // =======
    
    // Get function to retrieve the nth parameter.
    template<int m>
    typename boost::add_reference<typename nth_parameter_type<function_parameter_tuple<Head, Tail>, m>::ret>::type get_parameter()
    { return get_parameter_internal(nth_parameter_tag<m>()); };

    // Get functions to retrieve the element and the
    // tail. These must be public for the copy constructors.
    element_const_reference_type get_element() const
    { return m_element; }
    //
    rest_const_reference_type get_rest() const
    { return m_rest; }
    
  private:
    // First element and nested object containing the remaining
    // elements.
    element_reference_type m_element;
    nested_type m_rest;

    // Get function to retrieve the nth parameter, default implementation:
    // look in the nested rest object.
    template<int m>
    typename boost::add_reference<typename nth_parameter_type<function_parameter_tuple<Head, Tail>, m>::ret>::type get_parameter_internal(nth_parameter_tag<m>)
    { return m_rest.get_parameter<m-1>(); };

    // Get function to retrieve the nth parameter, overload for m=0:
    // return the element.
    element_reference_type get_parameter_internal(nth_parameter_tag<0>)
    { return m_element; };
    
    // Assigning and copying these babies is bad news.
    function_parameter_tuple(const function_parameter_tuple&);
    function_parameter_tuple& operator=(const function_parameter_tuple&);

  };

  // function_parameter_tuple type generator
  // =======================================
  
  template<
   typename T1=function_nil, 
   typename T2=function_nil, 
   typename T3=function_nil, 
   typename T4=function_nil, 
   typename T5=function_nil, 
   typename T6=function_nil, 
   typename T7=function_nil, 
   typename T8=function_nil, 
   typename T9=function_nil, 
   typename T10=function_nil, 
   typename T11=function_nil, 
   typename T12=function_nil, 
   typename T13=function_nil, 
   typename T14=function_nil, 
   typename T15=function_nil, 
   typename T16=function_nil,
   typename T17=function_nil
   /* EXTEND */
  >
  class make_function_parameter_tuple_type
  {
  public:
    typedef typename
      function_parameter_tuple<T1, typename make_function_parameter_tuple_type<
      T2, 
      T3, 
      T4, 
      T5, 
      T6, 
      T7, 
      T8, 
      T9, 
      T10, 
      T11, 
      T12, 
      T13, 
      T14, 
      T15, 
      T16, 
      T17 
      /* EXTEND */ 
      >::ret> ret;
  };

  // Kicks in for zero arguments passed
  template<>
  class make_function_parameter_tuple_type<
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil, 
    function_nil,
    function_nil, 
    function_nil
    /* EXTEND */
    >
 {
    public:
      typedef typename function_nil ret;
  };

}

#endif