#include "mojo.h"
#include <iostream>
#include <cassert>
using namespace std;

class Y : public mojo::enabled<Y> 
{
public:
    Y() // default constructor
    {
        cout << "CREATING default-constructed object at " << this 
            << '\n';
    }
    
    Y(const Y& rhs) // source is a const value
    {
        cout << "COPYING const value at " << &rhs 
            << " into object at " << this << '\n';
    }
    
    Y(mojo::fnresult<Y> src) // source is a fnresult
    {
        Y& rhs = src;
        cout << "MOVING mojo::fnresult<Y> at " << &rhs 
            << " into object at " << this << '\n';
    }

    Y(mojo::temporary<Y> src) // source is a temporary
    {
        Y& rhs = src.get();
        cout << "MOVING mojo::temporary<Y> at " << &rhs 
            << " into object at " << this << '\n';
    }

    ~Y()
    {
        //cout << "DESTROYING object at " << this << '\n';
    }
};

const Y MakeConstY()
{
    return Y();
}

mojo::fnresult<Y> MakeY()
{
    //if (cout) return Y();
    Y x;
    return x;
}

void TakeConstY(const Y&)
{
}

void TakeY(Y&)
{
}

void Discriminate(mojo::temporary<Y>) {}
void Discriminate(mojo::constant<Y>) {}
void Discriminate(Y& obj) { Discriminate(mojo::constant<Y>(obj)); }


////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

int main()
{
    Y nonConstLValue;
    const Y constLValue;
    
    //Y y1(constLValue);
    //Y y2(nonConstLValue);
    //Y y3(MakeConstY());
    Y y4(MakeY());

    TakeConstY(Y());

    Discriminate(Y()); // calls Discriminate(mojo::temporary<Y>)
    Discriminate(MakeY()); // calls Discriminate(mojo::temporary<Y>)
    Discriminate(constLValue); // calls Discriminate(mojo::constant<Y>)
    Discriminate(nonConstLValue); // calls Discriminate(Y&)
    //TakeConstY(MakeY());

    //TakeY(Y());
    //TakeY(MakeY());
}